#include <main.hpp>

/************************************************
 *              Global Variables                *
 ************************************************/
RemoteVariable aqi = RemoteVariable();
ESPMegaPRO espmega = ESPMegaPRO();
const char *mode_names[] = AC_MODE_NAMES;
const char *fan_speed_names[] = AC_FAN_SPEED_NAMES;
AirConditioner ac_config = {
    .max_temperature = AC_MAX_TEMP,
    .min_temperature = AC_MIN_TEMP,
    .modes = AC_MODES,
    .mode_names = mode_names,
    .fan_speeds = AC_FAN_SPEEDS,
    .fan_speed_names = fan_speed_names,
    .getInfraredCode = &getInfraredCode};

ClimateCard ac = ClimateCard(AC_IR_TX_PIN, ac_config, AC_SENSOR_TYPE, AC_SENSOR_PIN, AC_IR_CHANNEL);

cud_display_conf_t cud_display_conf = {
    .aqi = &aqi,
    .uart = &DISPLAY_UART,
    .tx = DISPLAY_TX,
    .rx = DISPLAY_RX,
    .communication_baudrate = DISPLAY_COMMUNICATION_BAUDRATE,
    .ota_baudrate = DISPLAY_OTA_BAUDRATE,
    .light_pins = {LIGHT_ROW_1_PIN, LIGHT_ROW_2_PIN, LIGHT_ROW_3_PIN, LIGHT_ROW_4_PIN},
    .fan_pins = {FAN_ROW_1_PIN, FAN_ROW_2_PIN, FAN_ROW_3_PIN},
    .air_purifier_pin = AIR_PURIFIER_PIN,
    .mosquito_zapper_pin = MOSQUITO_ZAPPER_PIN,
    .socket_contactor_pin = SOCKET_CONTACTOR_PIN};

CUDDisplay cudDisplay = CUDDisplay(&cud_display_conf);

ESPMegaDisplayOTA cudDisplayOTA = ESPMegaDisplayOTA();
ESPMegaDisplayOTA internalDisplayOTA = ESPMegaDisplayOTA();

// iBeacon configuration
static esp_ble_adv_params_t ble_adv_params = {
    .adv_int_min = 0x20,
    .adv_int_max = 0x40,
    .adv_type = ADV_TYPE_NONCONN_IND,
    .own_addr_type = BLE_ADDR_TYPE_PUBLIC,
    .channel_map = ADV_CHNL_ALL,
    .adv_filter_policy = ADV_FILTER_ALLOW_SCAN_ANY_CON_ANY,
};

esp_ble_ibeacon_vendor_t vendor_config_custom;

/************************************************
 *            End of Global Variables           *
 ************************************************/

/**
 * Generate a unique UUID based on ESP32's chip ID
 * @param uuid Output buffer for the UUID (16 bytes)
 */
void generate_unique_uuid(uint8_t *uuid) {
  // Get ESP32 chip ID (6 bytes)
  uint8_t mac[6];
  esp_efuse_mac_get_default(mac);
  
  // Use a fixed prefix (8 bytes)
  static const uint8_t prefix[8] = {0xFD, 0xA5, 0x06, 0x93, 0xA4, 0xE2, 0x4F, 0xB1};
  
  // Copy prefix to first 8 bytes of UUID
  memcpy(uuid, prefix, 8);
  
  // Copy MAC address to next 6 bytes
  memcpy(uuid + 8, mac, 6);
  
  // Fill the remaining 2 bytes with a hash of the MAC
  uint16_t hash = mac[0] + (mac[1] << 8) + mac[2] + (mac[3] << 8) + mac[4] + (mac[5] << 8);
  uuid[14] = hash & 0xFF;
  uuid[15] = (hash >> 8) & 0xFF;
}

void setup()
{
  // Initialize both displayports
  INTERNAL_DISPLAY_UART.begin(INTERNAL_DISPLAY_BAUDRATE);
  DISPLAY_UART.begin(DISPLAY_COMMUNICATION_BAUDRATE, SERIAL_8N1, DISPLAY_RX, DISPLAY_TX);
  ESP_LOGI("CUD IoT OS", "Starting Initialization");
  // Reset both display
  ESP_LOGV("CUD IoT OS", "Resetting both displays");
  send_stop_bytes(INTERNAL_DISPLAY_UART);
  INTERNAL_DISPLAY_UART.print("rest");
  send_stop_bytes(INTERNAL_DISPLAY_UART);
  send_stop_bytes(DISPLAY_UART);
  DISPLAY_UART.print("rest");
  send_stop_bytes(DISPLAY_UART);
  // Wait for the display to reset
  vTaskDelay(1000 / portTICK_PERIOD_MS);
  // Initialize the ESPMegaPRO
  ESP_LOGV("CUD IoT OS", "Initializing ESPMegaPRO");
  espmega.begin();
  // Set Pin 0 - 15 value to 4095
  for (uint8_t i = 0; i < 16; i++)
    espmega.outputs.setValue(i,4095);
  
  // Initialize IoT Modules
  ESP_LOGV("CUD IoT OS", "Initializing IoT Modules");
  espmega.setTimezone("ICT-7");
  espmega.enableIotModule();
  // Setup Ethernet
  ESP_LOGV("CUD IoT OS", "Initializing Ethernet");
  ETH.begin();
  ESP_LOGD("CUD IoT OS", "Binding Ethernet Interface");
  espmega.iot->bindEthernetInterface(&ETH);
  // Connect to Network
  ESP_LOGV("CUD IoT OS", "Connecting to Network");
  espmega.iot->loadNetworkConfig();
  espmega.iot->connectNetwork();
  // Connect to the MQTT Broker
  ESP_LOGV("CUD IoT OS", "Connecting to MQTT Broker");
  espmega.iot->loadMqttConfig();
  vTaskDelay(3000 / portTICK_PERIOD_MS); // Wait for the network to stabilize
  espmega.iot->connectToMqtt();
  // Enable Web Server
  ESP_LOGV("CUD IoT OS", "Enabling Web Server");
  espmega.enableWebServer(80);
  // Initialize Internal Display
  ESP_LOGV("CUD IoT OS", "Initializing Internal Display");
  espmega.enableInternalDisplay(&INTERNAL_DISPLAY_UART);
  // Initialize Air Conditioner
  ESP_LOGV("CUD IoT OS", "Initializing Air Conditioner");
  espmega.installCard(2, &ac);
  ac.bindFRAM(&espmega.fram, AC_FRAM_ADDR);
  ac.loadStateFromFRAM();
  ac.setFRAMAutoSave(true);
  ESP_LOGD("CUD IoT OS", "Loaded Air Conditioner State from FRAM, Temperature: %d, Mode: %d, Fan Speed: %d", ac.getTemperature(), ac.getMode(), ac.getFanSpeed());
  espmega.display->bindClimateCard(&ac); // Show our climate card on the display
  // Register all cards with iot
  ESP_LOGV("CUD IoT OS", "Registering Cards with IoT");
  espmega.iot->registerCard(0);
  espmega.iot->registerCard(1);
  espmega.iot->registerCard(2);
  // Initialize Remote Variables
  ESP_LOGV("CUD IoT OS", "Initializing Remote Variables");
  aqi.begin(10, AQI_STATE_TOPIC, espmega.iot, true, AQI_REQUEST_TOPIC);
  // Input callbacks
  // This pre-load the input buffers
  // We need to do this to prevent switches that are left on the "on" position from triggering the callback
  ESP_LOGV("CUD IoT OS", "Pre-loading input buffers");
  espmega.inputs.preloadInputBuffer();
  espmega.inputs.registerCallback(handle_input_change);
  // Start the display routine
  ESP_LOGV("CUD IoT OS", "Starting CUDDisplay");
  cud_display_cards_t cards = {
      .inputCard = &espmega.inputs,
      .outputCard = &espmega.outputs,
      .ac = &ac,
      .fram = &espmega.fram,
      .iot = espmega.iot,
  };
  cudDisplay.begin(cards);
  espmega.iot->registerMqttCallback(handle_mqtt_message);
  cudDisplayOTA.begin("/cuddisp", &cudDisplay, espmega.webServer);
  internalDisplayOTA.begin("/intdisp", espmega.display, espmega.webServer);
  
  // Initialize iBeacon with unique UUID
  ESP_LOGV("CUD IoT OS", "Initializing iBeacon");
  esp_bt_controller_config_t bt_cfg = BT_CONTROLLER_INIT_CONFIG_DEFAULT();
  esp_bt_controller_init(&bt_cfg);
  esp_bt_controller_enable(ESP_BT_MODE_BLE);
  esp_bluedroid_init();
  esp_bluedroid_enable();
  esp_ble_gap_register_callback(esp_gap_cb);
  
  // Generate unique UUID for this device
  uint8_t uuid[16];
  generate_unique_uuid(uuid);
  
  // Configure iBeacon with unique UUID
  memcpy(vendor_config_custom.proximity_uuid, uuid, sizeof(uuid));
  vendor_config_custom.major = ENDIAN_CHANGE_U16(ESP_MAJOR);
  vendor_config_custom.minor = ENDIAN_CHANGE_U16(ESP_MINOR);
  vendor_config_custom.measured_power = static_cast<int8_t>(0xC5);
  
  esp_ble_ibeacon_t ibeacon_adv_data;
  esp_ble_config_ibeacon_data(&vendor_config_custom, &ibeacon_adv_data);
  esp_ble_gap_config_adv_data_raw((uint8_t *)&ibeacon_adv_data, sizeof(ibeacon_adv_data));
  
  #ifdef TAMPER_DETECTION
  espmega.display->registerPageChangeCallback(pageChangeCallback);
  espmega.display->registerTouchCallback(touchCallback);
  #endif
  ESP_LOGI("CUD IoT OS", "Initialization Complete");
}

void loop()
{
  espmega.loop();
  cudDisplay.loop();
}

void send_stop_bytes(HardwareSerial &uart)
{
  uart.write(0xFF);
  uart.write(0xFF);
  uart.write(0xFF);
}

void handle_input_change(uint8_t pin, bool state)
{
  // When input pin 0 - 6 changes, toggle the corresponding output pin
  if (pin >= 0 && pin <= 6)
  {
    espmega.outputs.setState(pin, !espmega.outputs.getState(pin));
  }
}

void handle_mqtt_message(char *topic, char *payload)
{
  ESP_LOGD("CUD IoT OS", "MQTT Message Received: %s, %s", topic, payload);
}

static void esp_gap_cb(esp_gap_ble_cb_event_t event, esp_ble_gap_cb_param_t *param)
{
  esp_err_t err;

  switch (event) {
    case ESP_GAP_BLE_ADV_DATA_RAW_SET_COMPLETE_EVT: {
      esp_ble_gap_start_advertising(&ble_adv_params);
      break;
    }
    case ESP_GAP_BLE_ADV_START_COMPLETE_EVT:
      // adv start complete event to indicate adv start successfully or failed
      if ((err = param->adv_start_cmpl.status) != ESP_BT_STATUS_SUCCESS) {
        ESP_LOGE("iBeacon", "Advertising start failed, error %s", esp_err_to_name(err));
      } else {
        ESP_LOGI("iBeacon", "Advertising start successfully");
      }
      break;
    case ESP_GAP_BLE_ADV_STOP_COMPLETE_EVT:
      if ((err = param->adv_stop_cmpl.status) != ESP_BT_STATUS_SUCCESS) {
        ESP_LOGE("iBeacon", "Advertising stop failed, error %s", esp_err_to_name(err));
      } else {
        ESP_LOGI("iBeacon", "Advertising stop successfully");
      }
      break;
    default:
      break;
  }
}

#ifdef TAMPER_DETECTION
void pageChangeCallback(uint8_t page)
{
  char buffer[10];
  sprintf(buffer, "%d", page);
  espmega.iot->publishRelative("/tamper/page", buffer);
}
void touchCallback(uint8_t page, uint8_t component, uint8_t touchType)
{
  char buffer[20];
  sprintf(buffer, "%d,%d,%d", page, component, touchType);
  espmega.iot->publishRelative("/tamper/touch", buffer);
}
#endif